#ifndef ALIGNMENT_MATRIX
#define ALIGNMENT_MATRIX

#define ALN_MATRIX_BOUNDS_CHECK

#include <exception>
#include <iostream>

class index_out_of_bounds_exception : public std::exception
{
	virtual std::string what()
	{
		return std::string("internal error in alignment matrix: index out of bounds.\n");
	}
};

/// provide elementary access to a matrix in an optimized way.
/// if the pragma variable ALN_MATRIX_BOUNDS_CHECK is defined,
/// bounds will be checked whenever possible. If left undefined, no
/// bounds will be checked at all.
template <class score_type>
class alignment_matrix
{
	
public:
	
	/// number of rows
	int M;
	/// number of columns
	int N;

	/// matrix entries
	score_type* buf;
		
	alignment_matrix();

	~alignment_matrix();

	alignment_matrix(int M_, int N_);

	void init(int M_, int N_);
	
	inline score_type& operator()(const int& i, const int& j);

};

// =========================================================
template <class score_type>
void alignment_matrix<score_type>::init(int M_, int N_)
// =========================================================
{

	if (buf!=NULL)
		delete[] buf;
	
#ifdef ALN_MATRIX_BOUNDS_CHECK
	if (M_< 0 || N_<0)
		throw index_out_of_bounds_exception();
#endif
	M = M_;
	N = N_;
	buf = new score_type[(M+2)*(N+2)+2];
}

// =========================================================
template <class score_type>
alignment_matrix<score_type>::alignment_matrix()
// =========================================================
{
	buf = NULL;
	M = 0;
	N = 0;
}

// =========================================================
template <class score_type>
alignment_matrix<score_type>::~alignment_matrix()
// =========================================================
{
#ifdef ALN_MATRIX_BOUNDS_CHECK
	//if (buf==NULL)
	//	throw index_out_of_bounds_exception();
#endif
	if (buf!=NULL)
		delete[] buf;
	buf = NULL;
	
}

// =========================================================
template <class score_type>
alignment_matrix<score_type>::alignment_matrix(int M_, int N_)
// =========================================================
{
#ifdef ALN_MATRIX_BOUNDS_CHECK
	if (M_< 0 || N_<0)
		throw index_out_of_bounds_exception();
#endif
	M = M_;
	N = N_;
	buf = new score_type[(M+1)*(N+1)+1];
	if (buf==NULL)
		throw index_out_of_bounds_exception();
}

// =========================================================
template <class score_type>
inline score_type& alignment_matrix<score_type>::operator()(const int& i, const int& j)
// =========================================================
{
#ifdef ALN_MATRIX_BOUNDS_CHECK
	if (i<0 || j<0 || i>M || j>N)
	{
		throw index_out_of_bounds_exception();
	}
#endif
	return buf[(N+1)*i+j];
}



#endif


