#include "average_linkage_cluster.h"

typedef alignment_matrix<int> matrix_type;

typedef average_linkage_cluster::point point;
typedef average_linkage_cluster::hierarchy hierarchy;
typedef average_linkage_cluster::point_vector point_vector;
typedef average_linkage_cluster::int_vector int_vector;
typedef average_linkage_cluster::class_list class_list;
typedef average_linkage_cluster::int_vector_list int_vector_list;
typedef average_linkage_cluster::hierarchy_iterator hierarchy_iterator;
typedef average_linkage_cluster::class_list_iterator class_list_iterator;
typedef average_linkage_cluster::int_vector_list_iterator int_vector_list_iterator;
typedef int score_type;

score_type hierarchy_node::gap_open_penalty = 100;

score_type hierarchy_node::gap_ext_penalty = 10;

bool hierarchy_node::gap_cost_changed = false;

bool average_linkage_cluster::verbose = false;

bool average_linkage_cluster::clustalw_output = false;

// hierarchy_node

// =========================================================
template<class X>
X abs(X val)
// =========================================================
{
	if (val>=0)
		return val;
	else
		return -val;
}

// =========================================================
bool hierarchy_node::is_leaf()
// =========================================================
{return true;};

// =========================================================
void hierarchy_node::print(std::ostream& o) {o<<"VOID";}
// =========================================================

// =========================================================
float hierarchy_node::merge(point& result, bool weighted)
// =========================================================
{
	return 0.;
}

// =========================================================
void hierarchy_node::set_weight(float parent_score, float parent_weight)
// =========================================================
{weight = 1.; score = 0.;}

// =========================================================
float hierarchy_node::get_score() const
// =========================================================
{return score;} 

// =========================================================
float hierarchy_node::get_max_leaf_weight() 
// =========================================================
{return 0.;} 

// =========================================================
void hierarchy_node::normalize_leaf_weight(float wght_max) 
// =========================================================
{} 

// =========================================================
hierarchy_node::hierarchy_node()
// =========================================================
{score = 0.;}

// =========================================================
hierarchy_node::hierarchy_node(const point& p)
// =========================================================
{score = 0.; P = p;}

// =========================================================
hierarchy_node::~hierarchy_node()
// =========================================================
{}

	
// =========================================================
internal_hierarchy_node::internal_hierarchy_node(hierarchy_node* node_1, hierarchy_node* node_2, const point& representative_, float score_)
// =========================================================
{
	left = node_1;
	right = node_2;
	P = representative_;
	score = score_;
	num_of_points = left->num_of_points + right->num_of_points;
}

// =========================================================
bool internal_hierarchy_node::is_leaf()
// =========================================================
{return false;}

// =========================================================
float internal_hierarchy_node::get_max_leaf_weight()
// =========================================================
{
	float wght_lft = left->get_max_leaf_weight();
	float wght_rgt = right->get_max_leaf_weight();
	if (wght_lft > wght_rgt)
		return wght_lft;
	return wght_rgt;
		
}

// =========================================================
void internal_hierarchy_node::normalize_leaf_weight(float wght_max) 
// =========================================================
{
	left->normalize_leaf_weight(wght_max);
	right->normalize_leaf_weight(wght_max);
} 


// =========================================================
void internal_hierarchy_node::print(std::ostream& o)
// =========================================================
{
	o<<"( ";
	left->print(o);
	o<<" , ";
	right->print(o);
	o<<" )";
}

// =========================================================
float internal_hierarchy_node::merge(point& result, bool weighted)
// =========================================================
{
	point left_aln, right_aln;
	float left_score = left->merge(left_aln,weighted);
	float rght_score = right->merge(right_aln,weighted);
	float score = ::merge(left_aln,right_aln,result,weighted,gap_open_penalty,gap_ext_penalty,gap_cost_changed);
	P = result;
	int L_K, R_K;
	
	if (weighted)
	{
		// adjust distances in the new best alignment 
		float dist = distance(left_aln,right_aln);
		float_list new_weights;
		new_weights.clear();
		float_list weights_L = left_aln.get_weights();
		float_list::iterator w_it = weights_L.begin();
		float_list::iterator w_end = weights_L.end();
		L_K = left_aln.K;
		for (; w_it!=w_end; ++w_it)
		{
			float new_weight;
			if (L_K==1)
				new_weight = dist / 2.;
			else
				new_weight = (*w_it+dist)/(2.*(float)L_K);
			new_weights.push_back(new_weight);
		}
		float_list weights_R = right_aln.get_weights();
		w_it = weights_R.begin();
		w_end = weights_R.end();
		R_K = right_aln.K;
		for (; w_it!=w_end; ++w_it)
		{
			float new_weight;
			if (R_K==1)
				new_weight = dist / 2.;
			else
				new_weight = (*w_it+dist)/(2.*(float)R_K);
			new_weights.push_back(new_weight);
		}
		P.set_weights(new_weights);
 
		//if (verbose)
		//{
		//	std::cout<<"weights after add: ";
		//	float_list line_weights = P.line_weights;	
		//	float_list::iterator Yl_it = line_weights.begin();
		//	float_list::iterator Yl_end = line_weights.end();
		//	for (; Yl_it!=Yl_end; ++Yl_it)
		//	{
		//		std::cout<<*Yl_it<<"; ";
		//		float dbx = *Yl_it;
		//	}
		//	std::cout<<"\n";
		//}
	}
	
	return score;
}


// =========================================================
void internal_hierarchy_node::set_weight(float parent_score, float parent_weight)
// =========================================================
{
	if (score!=0.)
		weight = parent_weight + abs((score - parent_score) / (score*(float)num_of_points));
	else
		weight = parent_weight;
	left->set_weight(score,weight);
	right->set_weight(score,weight);
}

// =========================================================
internal_hierarchy_node::~internal_hierarchy_node()
// =========================================================
{
	delete left;
	delete right;
}

// =========================================================
bool hierarchy_leaf::is_leaf()
// =========================================================
{return true;}
	
// =========================================================
float hierarchy_leaf::get_max_leaf_weight()
// =========================================================
{return weight;}

// =========================================================
void hierarchy_leaf::normalize_leaf_weight(float wght_max) 
// =========================================================
{
	if (wght_max!=0.)
		weight /= wght_max;
} 


// =========================================================
hierarchy_leaf::hierarchy_leaf()
// =========================================================
{
	score = 0.;
	weight = 1.;
	num_of_points = 1;
}
	
// =========================================================
hierarchy_leaf::hierarchy_leaf(const point& p)
// =========================================================
{
	num_of_points = 1;
	score = 0.;
	weight = 1.;
	P = p;
}
	
// =========================================================
hierarchy_leaf::hierarchy_leaf(const point& p, float s)
// =========================================================
{
	num_of_points = 1;
	score = s;
	weight = 1.;
	P = p;
}

// =========================================================
void hierarchy_leaf::print(std::ostream& o)
// =========================================================
{
	o<<*(P.line_names.begin());
}

// =========================================================
float hierarchy_leaf::merge(point& result, bool weighted)
// =========================================================
{
	result = P;
	return 0;
}

// =========================================================
void hierarchy_leaf::set_weight(float parent_score, float parent_weight)
// =========================================================
{
	if (score!=0.)
		weight = parent_weight + abs((score - parent_score) / score);
	else
		weight = parent_weight;
	P.set_weights(weight);
}

// =========================================================
hierarchy_leaf::~hierarchy_leaf()
// =========================================================
{}

// =========================================================
void print_class(std::vector<point> v)
// =========================================================	  
{
	average_linkage_cluster::point_vector::iterator v_it = v.begin();
	average_linkage_cluster::point_vector::iterator v_end = v.end();
	for (; v_it!=v_end; ++v_it)
	{
		std::cout<<*v_it<<"; ";
	}	
}

// =========================================================
void average_linkage_cluster::set_weighted(bool w)
// =========================================================
{
	weighted = w;
}

// =========================================================
void average_linkage_cluster::print(std::ostream& o)
// =========================================================
{
	root->print(o);
	o<<"\n";
}
	

// =========================================================
average_linkage_cluster::average_linkage_cluster()
// =========================================================
{
	root = NULL;
	root_score = FLT_MIN;
	header_width = 8;
	body_width = 60;
	points.clear();
	gap_cost_changed = false;
	weighted = false;
}

// =========================================================
average_linkage_cluster::~average_linkage_cluster()
// =========================================================
{
	if (root!=NULL)
	{
		delete root;
	}
}


// =========================================================
average_linkage_cluster::average_linkage_cluster(
	const average_linkage_cluster::point_vector &new_points
	)
// =========================================================
{
	root = NULL;
	root_score = FLT_MIN;
	init(new_points);
	weighted = false;
	header_width = 8;
	body_width = 60;
	gap_cost_changed = false;
}

// =========================================================
void average_linkage_cluster::init(
	const average_linkage_cluster::point_vector &new_points
	)
// =========================================================
{
	clear();
	points = new_points;
	N = points.size();
}

// =========================================================
void average_linkage_cluster::clear()
// =========================================================
{
	
	if (root!=NULL)
	{
		root = NULL;
		delete root;
	}
	
}

// =========================================================
void average_linkage_cluster::run()
// =========================================================
{
	
	hierarchy roots;
	
	// dummy instance for computing "alignment-scores" of each
	// points[i] object; evtl. fix this in future by making match score
	// variables static in the pairwise alignment class...
	pairwise_cyclic_alignment A;
	if (gap_cost_changed)
		A.set_gap_costs(gap_open_penalty,gap_ext_penalty);
	int i;
	for (i=0; i<N; i++)
	{
		// use dummy object A (see above) to get a reasonable score for
		// points[i]
		hierarchy_leaf* leaf_i = new hierarchy_leaf(points[i],A.alignment_score(points[i]));
		roots.push_back(leaf_i);
		point_vector class_i(1);
	}

	hierarchy_iterator root_1, root_2, best_root_1, best_root_2;

	int size = N;
	point best_rep, rep;
	
	for (; size>1; --size)
	{
		root_1 = roots.begin();
		best_root_1 = roots.end();
		best_root_2 = roots.end();

		float best_score = -FLT_MAX;

		for (; root_1!=roots.end(); ++root_1)
		{
			root_2 = root_1;
			root_2++;
 			for (; root_2!=roots.end(); ++root_2)
			{
				float score = joint_score(*root_1,*root_2,rep);
				if (score>best_score)
				{
					best_score = score;
					best_root_1 = root_1;
					best_root_2 = root_2;
					best_rep = rep;
				}
			}
		}
		// adjust distances in the new best alignment 
		float dist = distance((*best_root_1)->P,(*best_root_2)->P);
		int L_K, R_K;
		L_K = (*best_root_1)->P.K;
		R_K = (*best_root_2)->P.K;
		float_list new_weights;
		new_weights.clear();
		float_list weights_L = (*best_root_1)->P.get_weights();
		float_list::iterator w_it = weights_L.begin();
		float_list::iterator w_end = weights_L.end();
		L_K = (*best_root_1)->P.K;
		for (; w_it!=w_end; ++w_it)
		{
			float new_weight;
			if (L_K==1)
				new_weight = dist / 2.;
			else
				new_weight = (*w_it+dist)/(2.*(float)L_K);
			new_weights.push_back(new_weight);
		}
		float_list weights_R = (*best_root_2)->P.get_weights();
		w_it = weights_R.begin();
		w_end = weights_R.end();
		R_K = (*best_root_2)->P.K;
		for (; w_it!=w_end; ++w_it)
		{
			float new_weight;
			if (R_K==1)
				new_weight = dist / 2.;
			else
				new_weight = (*w_it+dist)/(2.*(float)R_K);
			new_weights.push_back(new_weight);
		}
		best_rep.set_weights(new_weights);
		
		//std::cout<<"weights after add: ";
		//float_list line_weights = best_rep.line_weights;	
		//float_list::iterator Yl_it = line_weights.begin();
		//float_list::iterator Yl_end = line_weights.end();
		//for (; Yl_it!=Yl_end; ++Yl_it)
		//{
		//	std::cout<<*Yl_it<<"; ";
		//	float dbx = *Yl_it;
		//}
		//std::cout<<"\n";

		merge_hierarchies(best_root_1,best_root_2,roots,best_rep,best_score);
		
	}
	root = *roots.begin();
	if (verbose)
	{
		root->P.print_weight_string();
		std::cout<<"guide tree:\n";
		root->print(std::cout);
		std::cout<<"\n\n";
	}
	root_score = root->score;
}

// =========================================================
void average_linkage_cluster::run_tree_guided()
// =========================================================
{
	std::cout<<"setting up guide tree..."; std::cout.flush();
	setup_guide_tree();
	std::cout<<"done.\n";
	set_mid_point();
	compute_sequence_weights();
	std::cout<<"aligning along guide tree..."; std::cout.flush();
	merge_tree();
	std::cout<<"done.\n";
}

// =========================================================
void average_linkage_cluster::setup_guide_tree()
// =========================================================
{

	int i,j;
	score_matrix.init(N+1,N+1);
	hierarchy roots;
	int_vector_list classes;

	/// step 1: construct similarity matrix
	for (i=0; i<N; i++)
	{
		//score_matrix(i,i) = 0.;
		//score_matrix(i,i) = -FLT_MAX;
		for (j=i+1; j<N; j++)
		{
			point rep;
			point X = points[i];
			point Y = points[j];
			pairwise_cyclic_alignment A(X,Y);
			A.set_weighted(false);
			if (gap_cost_changed)
				A.set_gap_costs(gap_open_penalty,gap_ext_penalty);
			A.run_heuristic();
			int score = A.trace_score;
			score_matrix(i,j) = score;
			score_matrix(j,i) = score;
		}
	}

	// init trivial partition
	// using dummy object, see run()
	for (i=0; i<N; i++)
	{
		pairwise_cyclic_alignment A(points[i],points[i]);
		if (gap_cost_changed)
			A.set_gap_costs(gap_open_penalty,gap_ext_penalty);
		// using dummy object, see run()
		float score_i = A.alignment_score(points[i]);
		hierarchy_leaf* leaf_i = new hierarchy_leaf(points[i],score_i);
		roots.push_back(leaf_i);
		int_vector class_i(1);
		class_i[0] = i;
		classes.push_back(class_i);
	}

	/// step 2: set up guide tree
	hierarchy_iterator root_1, root_2, best_root_1, best_root_2;
	int_vector_list_iterator class_1, class_2, best_class_1, best_class_2;

	int size = N;
	point best_rep, rep;
	
	for (; size>1; --size)
	{
		root_1 = roots.begin();
		class_1 = classes.begin();
		best_root_1 = roots.end();
		best_root_2 = roots.end();
		best_class_1 = classes.end();
		best_class_2 = classes.end();

		float best_score = -FLT_MAX;
		
		for (; root_1!=roots.end(); ++root_1, ++class_1)
		{
			root_2 = root_1;
			class_2 = class_1;
			root_2++;
			class_2++;
			for (; root_2!=roots.end(); ++root_2, ++class_2)
			{
				float score = joint_avg_score(*class_1,*class_2,(*root_1)->get_score(),(*root_2)->get_score());
				if (score>best_score)
				{
					best_score = score;
					best_root_1 = root_1;
					best_root_2 = root_2;
					best_class_1 = class_1;
					best_class_2 = class_2;
				}
			}
		}
		merge_hierarchies(best_root_1,best_root_2,roots,best_score);
		merge_classes(best_class_1,best_class_2,classes);
		
	}
	root = *roots.begin();

}

// =========================================================
void average_linkage_cluster::set_mid_point()
// =========================================================
{
}
	
// =========================================================
void average_linkage_cluster::compute_sequence_weights()
// =========================================================
{
	root->set_weight(0,0.);
	float max_weight = root->get_max_leaf_weight();
	root->normalize_leaf_weight(max_weight);
}
	
// =========================================================
void average_linkage_cluster::merge_tree()
// =========================================================
{
	point result;
	if (gap_cost_changed)
	{
		hierarchy_node::gap_open_penalty = gap_open_penalty;
		hierarchy_node::gap_ext_penalty = gap_ext_penalty;
		hierarchy_node::gap_cost_changed = gap_cost_changed;
	}
	root_score = root->merge(result,weighted);
	if (verbose)
	{
		root->P.print_weight_string();
		std::cout<<"guide tree:\n";
		root->print(std::cout);
		std::cout<<"\n\n";
	}
	
}


// =========================================================
float average_linkage_cluster::joint_score(
	hierarchy_node* class_1,
	hierarchy_node* class_2,
	alignment& rep
	)
// =========================================================
{
	return merge(class_1->P,class_2->P, rep, weighted, gap_open_penalty, gap_ext_penalty, gap_cost_changed);
};


// =========================================================
void average_linkage_cluster::merge_hierarchies(
	average_linkage_cluster::hierarchy_iterator root_1,
	average_linkage_cluster::hierarchy_iterator root_2,
	average_linkage_cluster::hierarchy& roots,
	const point& new_representative,
	float score
	)
// =========================================================	  
{
	
	internal_hierarchy_node* node
		= new internal_hierarchy_node(*root_1,*root_2,new_representative,score);
	roots.erase(root_1);
	roots.erase(root_2);
	roots.push_back(node);
}

// =========================================================
void average_linkage_cluster::merge_hierarchies(average_linkage_cluster::hierarchy_iterator root_1, average_linkage_cluster::hierarchy_iterator root_2, average_linkage_cluster::hierarchy& roots, float score)
// =========================================================	  
{
	// representative for internal hierarchy node not needed here; can
	// be some arbitrary point p.
	point p;
	internal_hierarchy_node* node = new internal_hierarchy_node(*root_1,*root_2,p,score);
	roots.erase(root_1);
	roots.erase(root_2);
	roots.push_back(node);
}


// =========================================================
void average_linkage_cluster::merge_classes(
	average_linkage_cluster::class_list_iterator class_1,
	average_linkage_cluster::class_list_iterator class_2,
	class_list& classes
	)
// =========================================================	  
{
	// this implementation is still ugly and needs quite some improvement!
	point_vector new_class;
	
	point_vector::iterator cl_it = class_1->begin();
	point_vector::iterator cl_end = class_1->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	cl_it = class_2->begin();
	cl_end = class_2->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	classes.erase(class_1);
	classes.erase(class_2);
	classes.push_back(new_class);

}

// =========================================================
void average_linkage_cluster::merge_classes(
	average_linkage_cluster::int_vector_list_iterator class_1,
	average_linkage_cluster::int_vector_list_iterator class_2,
	average_linkage_cluster::int_vector_list& classes
	)
// =========================================================	  
{
	// this implementation is still ugly and needs quite some improvement!
	int_vector new_class;
	
	int_vector::iterator cl_it = class_1->begin();
	int_vector::iterator cl_end = class_1->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	cl_it = class_2->begin();
	cl_end = class_2->end();
	for (; cl_it != cl_end; ++cl_it)
		new_class.push_back(*cl_it);
	classes.erase(class_1);
	classes.erase(class_2);
	classes.push_back(new_class);

}

// =========================================================
float average_linkage_cluster::joint_avg_score(const int_vector& class_1, const int_vector& class_2, float class_1_score, float class_2_score)
// =========================================================
{
	int m = class_1.size();
	int n = class_2.size();
	int i,j;
	float avg_score = 0.0;
	
	for (i=0; i<m; i++)
		for (j=0; j<n; j++)
		{
			point class_1_i = points[class_1[i]];
			point class_2_j = points[class_2[j]];
			float score = score_matrix(class_1[i],class_2[j]);
			avg_score += score;
		}
	return (avg_score+class_1_score+class_2_score) / (m+n);
	
};

// =========================================================
void average_linkage_cluster::set_output_width(int header_w, int body_w)
// =========================================================
{
	header_width = header_w;
	body_width = body_w;
}

// =========================================================
void average_linkage_cluster::set_gap_costs(float open, float ext)
// =========================================================
{
	gap_cost_changed = true;
	gap_open_penalty = (int)(10.*open);
	gap_ext_penalty = (int)(10.*ext);
}

// =========================================================
std::string average_linkage_cluster::get_shifted_sequences()
// =========================================================
{
	std::string res = root->P.get_shifted_sequences();
	return res;
}



// =========================================================
std::ostream& operator<<(std::ostream& os, const average_linkage_cluster& A)
// =========================================================
{
	A.root->P.set_output_width(A.header_width, A.body_width);
	if (!average_linkage_cluster::clustalw_output)
		os<<"score: "<<A.root_score<<"\n\n"<<
			A.root->P<<"\n\n";
	else
		os<<"ClustalW format alignment produced by Cyclope; score: "<<A.root_score<<"\n\n"<<
			A.root->P<<"\n\n";
	
	return os;
}
