#ifndef AVERAGE_LINKAGE_CLUSTER
#define AVERAGE_LINKAGE_CLUSTER

#include "pairwise_cyclic_alignment.h"
#include "alignment.h"
#include "alignment_matrix.h"
#include <list>
#include <vector>
#include <limits.h>
#include <float.h>
#include <ostream>
#include <iostream>

typedef alignment_matrix<int> matrix_type;


// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//template<class point>
class hierarchy_node
{
public:

	typedef alignment point;
	typedef alignment::float_list float_list;
	//typedef pairwise_cyclic_alignment::score_type score_type;
	typedef int score_type;
	
	float score;

	float weight;

	point P;

	int num_of_points;

	static score_type gap_open_penalty;

	static score_type gap_ext_penalty;

	static bool gap_cost_changed;
	
	virtual bool is_leaf(); 

	virtual void print(std::ostream& o); 

	virtual float merge(point& result, bool weighted);

	virtual void set_weight(float parent_score, float parent_weight); 
	
	float get_score() const;

	virtual float get_max_leaf_weight();
	
	virtual void normalize_leaf_weight(float wght_max);
	
	hierarchy_node();
	
	hierarchy_node(const point& p);
	
	virtual ~hierarchy_node();

};
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++




// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class internal_hierarchy_node : public hierarchy_node
{
public:	

	hierarchy_node* left;
	hierarchy_node* right;
	
	internal_hierarchy_node(hierarchy_node* node_1, hierarchy_node* node_2, const point& representative_, float score_);

	bool is_leaf();

	void print(std::ostream& o);

	virtual float get_max_leaf_weight();
	
	virtual void normalize_leaf_weight(float wght_max);
	
	virtual float merge(point& result, bool weighted);

	virtual void set_weight(float parent_score, float parent_weight);
	
	~internal_hierarchy_node();
	
	
};
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++



// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class hierarchy_leaf : public hierarchy_node
{
	
public:

	bool is_leaf();
	
	hierarchy_leaf();
	
	hierarchy_leaf(const point& p);

	hierarchy_leaf(const point& p, float s);
	
	void print(std::ostream& o);

	virtual float get_max_leaf_weight();
	
	virtual void normalize_leaf_weight(float wght_max);
	
	virtual float merge(point& result, bool weighted);

	virtual void set_weight(float parent_score, float parent_weight);
		
	~hierarchy_leaf();
	
};
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++




// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
//template<class point>
class average_linkage_cluster
// =========================================================	
{
	
public:

	typedef alignment point;
	typedef  std::list< hierarchy_node* > hierarchy;
	typedef  std::vector<alignment> point_vector;
	typedef  std::vector<int> int_vector;
	typedef  std::list<point_vector> class_list;
	typedef  std::list<int_vector> int_vector_list;
	typedef  hierarchy::iterator hierarchy_iterator;
	typedef  class_list::iterator class_list_iterator;
	typedef  int_vector_list::iterator int_vector_list_iterator;
	typedef alignment::float_list float_list;
	
	/// number of points
	int N;
	
	/// points to be classified 
	point_vector points;

	matrix_type score_matrix;

	bool weighted;
			
	static bool verbose;

	static bool clustalw_output;
	
	/// The final outcome score of the guide-tree alignment obtained by
	/// pairwise alignments along the guide tree (if run_tree_guided()
	/// was applied; otherwise -- if run() was applied -- this value is not of
	/// any relevance.)
	float root_score;

	/// option for output format
	int header_width;

	/// option for output format
	int body_width;

	/// option for gap costs
	bool gap_cost_changed;

	/// option for gap costs
	int gap_open_penalty;

	/// option for gap costs
	int gap_ext_penalty;
	
	
	average_linkage_cluster();

	~average_linkage_cluster();

	average_linkage_cluster(const point_vector &new_points);

	hierarchy_node* root;

	void init(const point_vector &new_points);

	void clear();
	
	void run();

	void run_tree_guided();

	void setup_guide_tree();
	
	void set_mid_point();
	
	void compute_sequence_weights();
	
	void merge_tree();

	void print(std::ostream& o);

	void set_weighted(bool);
	
	float joint_score(hierarchy_node* class_rep_1, hierarchy_node* class_rep_2, alignment& rep);

	float joint_avg_score(const int_vector& class_1, const int_vector& class_2, float class_1_score, float class_2_score);
	
	static void merge_hierarchies(hierarchy_iterator root_1, hierarchy_iterator root_2, hierarchy& roots, const point& p, float score);

	static void merge_hierarchies(hierarchy_iterator root_1, hierarchy_iterator root_2, hierarchy& roots, float diam);
	
	static void merge_classes(class_list_iterator class_1, class_list_iterator class_2, class_list& classes);

	static void merge_classes(int_vector_list_iterator class_1, int_vector_list_iterator class_2, int_vector_list& classes);

	void set_output_width(int header_w, int body_w);

	void set_gap_costs(float open, float ext);

	std::string get_shifted_sequences();
	
	friend std::ostream& operator<<(std::ostream& os, const average_linkage_cluster& A);

};
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// +++++++++++++++++++++++++++++++++++++++++++++++++++++++++


#endif
