#include "pairwise_cyclic_alignment.h"
#define ALN_MATRIX_BOUNDS_CHECK
#include "alignment_matrix.h"
#include "limits.h"
#include "float.h"
#include "simple_local_alignment.h"

#include <algorithm>
#include <cstdio>
#include <cmath>

typedef pairwise_cyclic_alignment::score_type score_type;
typedef pairwise_cyclic_alignment::sequence_type sequence_type;

score_type pairwise_cyclic_alignment::SCORE_MIN = INT_MIN/2;
bool pairwise_cyclic_alignment::use_joint_match_score = false;
bool pairwise_cyclic_alignment::verbose = false;

// =========================================================
void _itoa(int val, std::string& buf, int base) {
// =========================================================
	
	int i;
	
	buf.reserve(64);
	buf = "";
	
	for(i=64; val>0 && i>0 ; --i, val /= base)
		buf = (char)((int)'0'+(val % base)) + buf;
	
}
	
// =========================================================
alignment_exception::alignment_exception(int id)
// =========================================================
{
	ID = id;
}

// =========================================================
std::string alignment_exception::what()
// =========================================================
{
	
	std::string str;
	
	//sprintf (buffer, "internal error in alignment procedure #%d\n", ID);
	std::string msg = std::string("internal error in alignment procedure");
 	return msg;
}



// =========================================================
float merge(alignment& aln_1, alignment& aln_2, pairwise_cyclic_alignment& result, bool weighted, score_type g_open, score_type g_ext, bool cost_change)
// =========================================================
{
	pairwise_cyclic_alignment A(aln_1,aln_2);
	if (cost_change)
		A.set_gap_costs(g_open, g_ext);
	A.set_weighted(weighted);
	A.run_heuristic();
	result = A;
	return A.trace_score;
}


// =========================================================
float merge(alignment& aln_1, alignment& aln_2, alignment& result, bool weighted, score_type g_open, score_type g_ext, bool cost_change)
// =========================================================
{
	pairwise_cyclic_alignment A(aln_1,aln_2);
	if (cost_change)
		A.set_gap_costs(g_open, g_ext);
	A.set_weighted(weighted);
	A.run_heuristic();
	result = A.result;
	return A.trace_score;
}

 // =========================================================
void shift(std::string& str, int k)
// =========================================================
{
	int n = str.size();
	int k_rep = k%n;
	k_rep = (k_rep>=0 ? k_rep : k_rep+n);
	
	std::string result = str.substr(k_rep,str.size()-k_rep) + str.substr(0,k_rep);
	str = result;
}

// =========================================================
void shift(alignment& aln, int k)
// =========================================================
{
	int i;
	int n = aln.size();
	int k_rep = k%n;
	k_rep = (k_rep>=0 ? k_rep : k_rep+n);
	
	alignment::string_vector buf(k_rep);
	for (i=0; i<k_rep; i++)
		buf[i] = aln[i];
	for (i=k_rep; i<n; i++)
		aln[i-k_rep] = aln[i];
	for (i=0; i<k_rep; i++)
		aln[i+n-k_rep] = buf[i];
}

// =========================================================
bool check_aln_vs_seq(const std::string& aln, const std::string& seq)
// =========================================================
{
	int i;
	std::string seq_aln = "";
	for (i=0; i<aln.size(); i++)
	{
		if (aln[i]!='-')
			seq_aln += aln[i];
	}
	return seq_aln==seq;
}

// =========================================================
score_type pairwise_cyclic_alignment::alignment_score(sequence_type aln1, sequence_type aln2)
// =========================================================
{
	score_type score = 0;
	if (aln1.size()!=aln2.size())
		return SCORE_MIN;
	int N = aln1.size();
	int i;
	for (i=0; i<N; i++)
	{
		if (!is_gap(aln1[i]) && !is_gap(aln2[i]))
			score += match_score(aln1[i],aln2[i]);
		else 
			if (is_gap(aln1[i]))
				if (i>0) 
					if (is_gap(aln1[i-1]))
						score += alpha;
					else
						score += gamma;
				else
					score += gamma;
			else // aln2[i]=='-'
				if (i>0) 
					if (is_gap(aln2[i-1]))
						score += alpha;
					else
						score += gamma;
				else
					score += gamma;
	}
	if (is_gap(aln1[0]) && is_gap(aln1[N-1]))
		score += alpha - gamma;
	if (is_gap(aln2[0]) && is_gap(aln2[N-1]))
		score += alpha - gamma;
	return score;
	
}

// =========================================================
score_type pairwise_cyclic_alignment::alignment_score(sequence_type aln)
// =========================================================
{
	score_type score = 0;
	int N = aln.size();
	int i;
	for (i=0; i<N; i++)
		if (!is_gap(aln[i]))
			score += match_score(aln[i],aln[i]);
	return score;
	
}

// =========================================================
inline score_type pairwise_cyclic_alignment::match_score(char x, char y)
// =========================================================
{
	if (x==y)
		return single_match_score;
	else
		return single_mismatch_score;
}

// =========================================================
inline score_type pairwise_cyclic_alignment::local_match_score(char x, char y)
// =========================================================
{
	if (x==y)
		return single_local_match_score;
	else
		return single_local_mismatch_score;
}

/*
// =========================================================
score_type pairwise_cyclic_alignment::match_score(const std::string& x, const std::string& y)
// =========================================================
{
	int mu = x.size();
	int nu = y.size();
	int freq_x[256];
	int freq_y[256];
	int* freq_x_A = &(freq_x['A']);
	int* freq_x_C = &(freq_x['C']);
	int* freq_x_G = &(freq_x['G']);
	int* freq_x_T = &(freq_x['T']);
	int* freq_x_U = &(freq_x['U']);
	int* freq_y_A = &(freq_x['A']);
	int* freq_y_C = &(freq_x['C']);
	int* freq_y_G = &(freq_x['G']);
	int* freq_y_T = &(freq_x['T']);
	int* freq_y_U = &(freq_x['U']);
	*freq_x_A = 0;
	*freq_x_C = 0;
	*freq_x_G = 0;
	*freq_x_T = 0;
	*freq_x_U = 0;
	*freq_y_A = 0;
	*freq_y_C = 0;
	*freq_y_G = 0;
	*freq_y_T = 0;
	*freq_y_U = 0;

	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it;
	score_type score = 0;
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
	{
		freq_x[*x_it]++;
	}
	for (y_it = y.begin(); y_it!=y.end(); ++y_it)
	{
		freq_y[*y_it]++;
	}
	score = (*freq_x_A)*(*freq_y_A) +(*freq_x_C)*(*freq_y_C) + (*freq_x_G)*(*freq_y_G) + (*freq_x_T)*(*freq_y_T) + (*freq_x_U)*(*freq_y_U);
	
	return (score_type)(((float)(single_ *score))/((float)(mu*nu)));
}*/

// =========================================================
score_type pairwise_cyclic_alignment::match_score(const std::string& x, const std::string& y, const float_list& w_x, const float_list& w_y)
// =========================================================
{
	int mu = x.size();
	int nu = y.size();
	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it;
	float_list::const_iterator f_it, g_it;
	
	score_type score = 0;
	for (x_it = x.begin(), f_it = w_x.begin(); x_it!=x.end(); ++x_it, ++f_it)
		for (y_it = y.begin(), g_it = w_y.begin(); y_it!=y.end(); ++y_it, ++g_it)
		{
			if (*x_it!='-' && *y_it!='-')
				score += (int)(match_score(*x_it,*y_it)*(*f_it)*(*g_it));
		}
	return (score_type)((float)score)/((float)(mu*nu));
}

// =========================================================
score_type pairwise_cyclic_alignment::match_score(const std::string& x, const std::string& y)
// =========================================================
{
	if (use_joint_match_score)
		return joint_match_score(x,y);
	int mu = x.size();
	int nu = y.size();
	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it;
	
	score_type score = 0;
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
		for (y_it = y.begin(); y_it!=y.end(); ++y_it)
		{
			if (*x_it!='-' && *y_it!='-')
				score += match_score(*x_it,*y_it);
		}
	//return (score_type)((float)score)/((float)(mu*nu));
	return (int)((float)score)/((float)(mu*nu));
}

// =========================================================
score_type pairwise_cyclic_alignment::joint_match_score(const std::string& x, const std::string& y)
// =========================================================
{
	int mu = x.size();
	int nu = y.size();
	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it;
	std::string::const_iterator xx_it, yy_it;
	
	score_type score = 0;
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
	{
		for (y_it = y.begin(); y_it!=y.end(); ++y_it)
		{
			if (*x_it!='-' && *y_it!='-')
				score += match_score(*x_it,*y_it);
		}
	}
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
	{
		xx_it = x_it;
		++xx_it;
		for (; xx_it!=x.end(); ++xx_it)
		{
			if (*x_it!='-' && *xx_it!='-')
				score += match_score(*x_it,*xx_it);
		}
	}
	for (y_it = y.begin(); y_it!=y.end(); ++y_it)
	{
		yy_it = y_it;
		++yy_it;
		for (; yy_it!=y.end(); ++yy_it)
		{
			if (*y_it!='-' && *yy_it!='-')
				score += match_score(*y_it,*yy_it);
		}
	}
	return (score_type)((float)score)/((float)(((mu+nu)*(mu+nu-1))/2));
}

// =========================================================
inline score_type pairwise_cyclic_alignment::local_match_score(std::string& x, std::string& y)
// =========================================================
{
	if (use_joint_match_score)
		return joint_local_match_score(x,y);
	int mu = x.size();
	int nu = y.size();
	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it;
	score_type score = 0;
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
		for (y_it = y.begin(); y_it!=y.end(); ++y_it)
			if (*x_it!='-' && *y_it!='-')
				score += local_match_score(*x_it,*y_it);
	return (score_type)((float)score)/((float)(mu*nu));
}

// =========================================================
inline score_type pairwise_cyclic_alignment::joint_local_match_score(const std::string& x, const std::string& y)
// =========================================================
{
	int mu = x.size();
	int nu = y.size();
	if (mu*nu==0)
		return 0;
	std::string::const_iterator x_it, y_it, xx_it, yy_it;
	score_type score = 0;
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
		for (y_it = y.begin(); y_it!=y.end(); ++y_it)
			if (*x_it!='-' && *y_it!='-')
				score += local_match_score(*x_it,*y_it);
	for (x_it = x.begin(); x_it!=x.end(); ++x_it)
	{
		xx_it = x_it;
		++xx_it;
		for (; xx_it!=x.end(); ++xx_it)
		{
			if (*x_it!='-' && *xx_it!='-')
				score += local_match_score(*x_it,*xx_it);
		}
	}
	for (y_it = y.begin(); y_it!=y.end(); ++y_it)
	{
		yy_it = y_it;
		++yy_it;
		for (; yy_it!=y.end(); ++yy_it)
		{
			if (*y_it!='-' && *yy_it!='-')
				score += local_match_score(*y_it,*yy_it);
		}
	}
	return (score_type)((float)score)/((float)(((mu+nu)*(mu+nu-1))/2));
}


// =========================================================
inline score_type trace_min(const score_type& a, const score_type& b, const score_type& c, int& trace)
// =========================================================
{
	if (a<b)
		if (a<c)	{
			trace = 0;
			return a;
		} else {
			trace = -1;
			return c;
		}
	else
		if (b<c)
		{
			trace = 1;
			return b;
		} else {
			trace = -1;
			return c;
		}			
}

// =========================================================
template<class score_type>
inline score_type min(const score_type& a, const score_type& b)
// =========================================================
{
	if (a<=b)
		return a;
	else
		return b;
}

// =========================================================
inline score_type trace_max(const score_type& a, const score_type& b, const score_type& c, int& trace)
// =========================================================
{
	if (a>b)
		if (a>c)	{
			trace = 0;
			return a;
		} else {
			trace = -1;
			return c;
		}
	else
		if (b>c)
		{
			trace = 1;
			return b;
		} else {
			trace = -1;
			return c;
		}			
}

// =========================================================
inline score_type max(const score_type& a, const score_type& b)
// =========================================================
{
	if (a>=b)
		return a;
	else
		return b;
}

// =========================================================
inline score_type max(const score_type& a, const score_type& b, const score_type& c)
// =========================================================
{
	return max(max(a,b),c);
}

// =========================================================
pairwise_cyclic_alignment::pairwise_cyclic_alignment()
// =========================================================
{
	weighted = false;
	header_width = 8;
	body_width = 60;
}

// =========================================================
//pairwise_cyclic_alignment::~pairwise_cyclic_alignment()
// =========================================================
//{
//}

// =========================================================
pairwise_cyclic_alignment::pairwise_cyclic_alignment(const sequence_type& X_, const sequence_type& Y_)
// =========================================================
{
	X = X_;
	Y = Y_;
	backup_X = X_;
	backup_Y = Y_;
	header_width = 8;
	body_width = 60;

	weighted = false;
	
	m = X.size();
	n = Y.size();

	init_matrices();
	trace_X = X_;
	trace_Y = Y_;

	// clustalw parameter set:
	//gamma = -10.;
	//alpha = -.1;
	//single_match_score = 1.9;
	//single_mismatch_score = 0.;
	
	gamma = -100;
	alpha = -1;
	single_match_score = 19;
	single_mismatch_score = 0;

	single_local_match_score = 14;
	single_local_mismatch_score = -5;
	local_gamma = -74;
	local_alpha = -1;

	// local_gamma = gamma * single_local_match_score / (single_local_match_score - single_local_mismatch_score)
	// local_alpha = max( 1 , alpha * single_local_match_score / (single_local_match_score - single_local_mismatch_score) )

}

// =========================================================
void pairwise_cyclic_alignment::set_weighted(bool w)
// =========================================================
{
	weighted = w;
}

// =========================================================
void pairwise_cyclic_alignment::run_heuristic()
// =========================================================
{

	int k,l,I,J;
	score_type score;
	
	init_local_start(0,0);
	compute_local_matrices(0,0);
	get_best_pos(I,J);
	score = A_H(I,J);
	trace_local(I,J);
	
	k = (best_run_start_X+best_run_end_X) / 2;
	l = (best_run_start_Y+best_run_end_Y) / 2;

	shift(X,k);
	shift(Y,l);
	
	init_match_start(0,0);
	if (!weighted) {
		compute_matrices(0,0);
	} else {
		compute_weighted_matrices(0,0);
	}
	
	score = A_H(m,n);
	if (!weighted)
		trace_sequences(0,0,false);
	else
		trace_weighted_sequences(0,0,false);
	if (trace_score!=score)
		throw alignment_exception(2001);
	result = trace_X;
	result.merge(trace_Y);
	
}


// =========================================================
void pairwise_cyclic_alignment::run()
// =========================================================
{
	int k;
	score_type best_score = SCORE_MIN;
	score_type score;
	int best_k = -1;
	bool gap_start;
	
	for (k=0; k<m; k++)
	{
		shift(X,1);
		init_match_start(0,0);
		compute_matrices(0,0);
		score = A_H(m,n);
		if (score>best_score)
		{
			best_score = score;
			best_k = k;
			trace_sequences(0, 0, false);
			gap_start = false;
		}
		
		init_gap_start(0,0);
		compute_matrices(0,0);
		score = A_Q(m,n);
		if (score>best_score)
		{
			best_score = score;
			best_k = k;
			trace_sequences(0, 0, true);
			gap_start = true;
		}
	}
}

// =========================================================
void pairwise_cyclic_alignment::run_linear()
// =========================================================
{
	//score_type best_score = SCORE_MIN;
	score_type score;
	//int best_k = -1;
	
	init_match_start(0,0);
	compute_matrices(0,0);
	score = A_H(m,n);
	trace_sequences(0, 0, false);
}

// =========================================================
void pairwise_cyclic_alignment::run_local()
// =========================================================
{

	int I,J;
	//score_type best_score = SCORE_MIN;
	score_type score;
	//int best_k = -1;
	
	init_local_start(0,0);
	compute_local_matrices(0,0);
	get_best_pos(I,J);
	score = A_H(I,J);
	trace_local(I,J);

}


// =========================================================
void pairwise_cyclic_alignment::trace_sequences(int k, int l, bool gap_start)
// =========================================================
{
	int i = m;
	int j = n;
	int s = 0;
	int ii,jj;
	//char cX, cY;
	trace_X.resize(m+n+1);
	trace_Y.resize(m+n+1);
	
	trace_score = 0;
	//score_type cs = A_H(m,n);
	int state;
	if (gap_start)
		state = -1;
	else
		trace_max(A_M(m,n),A_P(m,n),A_Q(m,n),state);	
	
	while (i>0 || j>0)
	{
		ii = (i-1+k)%m;
		jj = (j-1+l)%n;
		switch (state)
		{
			case 0:
				if (j<=0 || i<=0) {
					throw alignment_exception(1001);
				} else if (A_M(i,j)==A_M(i-1,j-1) + match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj]);
					state = 0;
					i--;j--;s++;
				} else if (A_M(i,j)==A_P(i-1,j-1) + match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj]);
					state = 1;
					i--;j--;s++;
			   } else if (A_M(i,j)==A_Q(i-1,j-1) + match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj]);
					state = -1;
					i--;j--;s++;		
				} else {
					throw alignment_exception(1002);					
				}
				break;
			case 1:
				if (i<=0) {
					throw alignment_exception(1003);
				} else if (A_P(i,j)==A_P(i-1,j)+alpha) {
					trace_X[s] = ii>=0 ? X[ii]: std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += alpha;
					state = 1;
					i--; s++;
				} else if (A_P(i,j)==A_M(i-1,j)+gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 0;
					i--; s++;
				} else if (A_P(i,j)==A_Q(i-1,j)+gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = -1;
					i--; s++;
				}
				break;
			case -1:
				if (j<=0) {
					throw alignment_exception(1004);
				} else if (A_Q(i,j)==A_Q(i,j-1)+alpha) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += alpha;
					state = -1;
					j--; s++;
				} else if (A_Q(i,j)==A_M(i,j-1)+gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 0;
					j--; s++;
				} else if (A_Q(i,j)==A_P(i,j-1)+gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 1;
					j--; s++;
				}
				break;
		}
	}
	trace_X.resize(s);
	trace_Y.resize(s);
	trace_X.L = s;
	trace_Y.L = s;
	trace_X.reverse();
	trace_Y.reverse();

}

// =========================================================
void pairwise_cyclic_alignment::trace_weighted_sequences(int k, int l, bool gap_start)
// =========================================================
{
	int i = m;
	int j = n;
	int s = 0;
	int ii,jj;
	//char cX, cY;
	trace_X.resize(m+n+1);
	trace_Y.resize(m+n+1);
	
	trace_score = 0;
	//score_type cs = A_H(m,n);
	int state;
	if (gap_start)
		state = -1;
	else
		trace_max(A_M(m,n),A_P(m,n),A_Q(m,n),state);	

	float_list X_normalized_line_weights = X.get_normalized_weights();
	float_list Y_normalized_line_weights = Y.get_normalized_weights();
	
	while (i>0 || j>0)
	{
		ii = (i-1+k)%m;
		jj = (j-1+l)%n;
		switch (state)
		{
			case 0:
				if (j<=0 || i<=0) {
					throw alignment_exception(1005);
				} else if (A_M(i,j)==A_M(i-1,j-1) + match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights)) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights);
					state = 0;
					i--;j--;s++;
				} else if (A_M(i,j)==A_P(i-1,j-1) + match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights)) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights);
					state = 1;
					i--;j--;s++;
			   } else if (A_M(i,j)==A_Q(i-1,j-1) + match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights)) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += match_score(X[ii],Y[jj],X_normalized_line_weights,Y_normalized_line_weights);
					state = -1;
					i--;j--;s++;		
				} else {
					throw alignment_exception(1006);
				}
				break;
			case 1:
				if (i<=0) {
					throw alignment_exception(1007);
				} else if (A_P(i,j)==A_P(i-1,j)+alpha) {
					trace_X[s] = ii>=0 ? X[ii]: std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += alpha;
					state = 1;
					i--; s++;
				} else if (A_P(i,j)==A_M(i-1,j)+gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 0;
					i--; s++;
				} else if (A_P(i,j)==A_Q(i-1,j)+gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = -1;
					i--; s++;
				}
				break;
			case -1:
				if (j<=0) {
					throw alignment_exception(1008);
				} else if (A_Q(i,j)==A_Q(i,j-1)+alpha) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += alpha;
					state = -1;
					j--; s++;
				} else if (A_Q(i,j)==A_M(i,j-1)+gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 0;
					j--; s++;
				} else if (A_Q(i,j)==A_P(i,j-1)+gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += gamma;
					state = 1;
					j--; s++;
				}
				break;
		}
	}
	trace_X.resize(s);
	trace_Y.resize(s);
	trace_X.L = s;
	trace_Y.L = s;
	trace_X.reverse();
	trace_Y.reverse();

}

// =========================================================
void pairwise_cyclic_alignment::get_best_pos(int& I, int& J)
// =========================================================
{
	int i,j;
	score_type best_score = A_M(1,1);
	I = 1;
	J = 1;
	for (i=0; i<=m; i++)
	{
		for (j=0; j<=n; j++)
		{
			if (A_M(i,j) > best_score)
			{
				best_score = A_M(i,j);
				I = i;
				J = j;
			}
		}
	}

}

/// I and J denote the end positions of the local alignment, typically
/// obtained by determining a local or global optimum in the (local)
/// alignment matrix.
// =========================================================
void pairwise_cyclic_alignment::trace_local(int I, int J)
// =========================================================
{
	local_aln_end_X = I;
	local_aln_end_Y = J;
	int i = I;
	int j = J;
	int s = 0;
	int ii,jj;
	//char cX, cY;
	trace_X.resize(m+n+1);
	trace_Y.resize(m+n+1);
	
	trace_score = 0;
	//score_type cs = A_H(m,n);

	// determine start state
	int state;
	//score_type s_H = A_H(I,J);
	score_type s_M = A_M(I,J);
	//score_type s_P = A_P(I,J);
	//score_type s_Q = A_Q(I,J);
	score_type s_M_m, s_P_m, s_Q_m, s_M_p, s_P_p, s_Q_p, s_M_q, s_P_q, s_Q_q, s_mtch;
	
	score_type run_score = 0;
	best_run_score = 0;
	int run_end_X = I;
	int run_end_Y = J;
	best_run_start_X = I-1;
	best_run_start_Y = J-1;
	best_run_end_X = I;
	best_run_end_Y = J;
	
	if (A_H(I,J)==0) {
		// this may happen if the sequences to be aligned do not have a character in common.
		return;
		//throw alignment_exception(1009);
	}
	if (A_H(I,J)==A_M(I,J)) {
		state = 0; 
	} else if (A_H(I,J)==A_P(I,J)) {
		state = 1;
	} else if (A_H(I,J)==A_Q(I,J)) {
		state = -1;
	} else {
		throw alignment_exception(1010);
		state = 2;
	}

	while (state!=2)
	{
		ii = (i-1)%m;
		jj = (j-1)%n;
		switch (state)
		{
			case 0:
				s_M = A_M(i,j);
				s_M_m = A_M(i-1,j-1);
				s_P_m = A_P(i-1,j-1);
				s_Q_m = A_Q(i-1,j-1);
				s_M_p = A_M(i-1,j);
				s_P_p = A_P(i-1,j);
				s_Q_p = A_Q(i-1,j);
				s_M_q = A_M(i,j-1);
				s_P_q = A_P(i,j-1);
				s_Q_q = A_Q(i,j-1);
				s_mtch = local_match_score(X[ii],Y[jj]);
				if (j<=0 || i<=0) {
					throw alignment_exception(1011);
				} else if (A_M(i,j)==A_M(i-1,j-1) + local_match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_match_score(X[ii],Y[jj]);
					run_score += local_match_score(X[ii],Y[jj]);
					state = 0;
					i--;j--;s++;
				} else if (A_M(i,j)==A_P(i-1,j-1) + local_match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_match_score(X[ii],Y[jj]);
					run_score += local_match_score(X[ii],Y[jj]);
					if (run_score > best_run_score) {
						best_run_score = run_score;
						best_run_end_X = run_end_X;
						best_run_end_Y = run_end_Y;
						best_run_start_X = ii;
						best_run_start_Y = jj;
					}
					state = 1;
					i--;j--;s++;
			   } else if (A_M(i,j)==A_Q(i-1,j-1) + local_match_score(X[ii],Y[jj])) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_match_score(X[ii],Y[jj]);
					run_score += local_match_score(X[ii],Y[jj]);
					if (run_score > best_run_score) {
						best_run_score = run_score;
						best_run_end_X = run_end_X;
						best_run_end_Y = run_end_Y;
						best_run_start_X = ii;
						best_run_start_Y = jj;
					}
					state = -1;
					i--;j--;s++;		
				} else {
					throw alignment_exception(1012);
				}
				if (A_M(i,j)==0) {
					local_aln_start_X = i+1;
					local_aln_start_Y = j+1;
					state = 2;
				}
				break;
			case 1:
				if (i<=0) {
					throw alignment_exception(1013);
				} else if (A_P(i,j)==A_P(i-1,j)+local_alpha) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += local_alpha;
					state = 1;
					i--; s++;
				} else if (A_P(i,j)==A_M(i-1,j)+local_gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += local_gamma;
					state = 0;
					i--; s++;
					run_score = 0;
					run_end_X = i;
					run_end_Y = j;
				} else if (A_P(i,j)==A_Q(i-1,j)+local_gamma) {
					trace_X[s] = ii>=0 ? X[ii] : std::string(trace_X.height(),'-');
					trace_Y[s] = std::string(trace_Y.height(),'-');
					trace_score += local_gamma;
					state = -1;
					i--; s++;
				} else {
					throw alignment_exception(1014);
				}
				break;
			case -1:
				if (j<=0) {
					throw alignment_exception(1015);
				} else if (A_Q(i,j)==A_Q(i,j-1)+local_alpha) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_alpha;
					state = -1;
					j--; s++;
				} else if (A_Q(i,j)==A_M(i,j-1)+local_gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_gamma;
					state = 0;
					j--; s++;
					run_score = 0;
					run_end_X = i;
					run_end_Y = j;
				} else if (A_Q(i,j)==A_P(i,j-1)+local_gamma) {
					trace_X[s] = std::string(trace_X.height(),'-');
					trace_Y[s] = jj>=0 ? Y[jj] : std::string(trace_Y.height(),'-');
					trace_score += local_gamma;
					state = 1;
					j--; s++;
				} else {
					throw alignment_exception(1016);
				}
					
				break;
		}
	}
	trace_X.resize(s);
	trace_Y.resize(s);
	trace_X.reverse();
	trace_Y.reverse();

}

// =========================================================
void pairwise_cyclic_alignment::init_matrices()
// =========================================================
{
	// init matrices
	A_M.init(m+1,n+1);
	A_H.init(m+1,n+1);
	A_P.init(m+1,n+1);
	A_Q.init(m+1,n+1);
}


// =========================================================
void pairwise_cyclic_alignment::init_match_start(int k, int l)
// =========================================================
{
	int i,j;
	A_M(0,0) = 0;
	A_H(0,0) = 0;
	A_P(0,0) = SCORE_MIN;
	A_Q(0,0) = SCORE_MIN;
	for (i=1; i<=m; i++)
	{
		A_P(i,0) = SCORE_MIN;
		A_Q(i,0) = SCORE_MIN;
		A_M(i,0) = SCORE_MIN;
		A_H(i,0) = SCORE_MIN;
	}
	for (j=1; j<=n; j++)
	{
		A_Q(0,j) = SCORE_MIN;
		A_P(0,j) = SCORE_MIN;
		A_M(0,j) = SCORE_MIN;
		A_H(0,j) = SCORE_MIN;
	}
}
	
// =========================================================
void pairwise_cyclic_alignment::init_gap_start(int k, int l)
// =========================================================
{
	int i,j;
	A_M(0,0) = SCORE_MIN;
	A_P(0,0) = SCORE_MIN;
	A_H(0,0) = 0;
	A_Q(0,0) = 0;
	for (i=1; i<=m; i++)
	{
		A_P(i,0) = SCORE_MIN;
		A_Q(i,0) = SCORE_MIN;
		A_M(i,0) = SCORE_MIN;
		A_H(i,0) = SCORE_MIN;
	}
	for (j=1; j<=n; j++)
	{
		A_P(0,j) = SCORE_MIN;
		A_Q(0,j) = A_Q(0,j-1) + alpha;
		A_M(0,j) = SCORE_MIN;
		A_H(0,j) = A_Q(0,j);
	}
	
}
	
// =========================================================
void pairwise_cyclic_alignment::init_local_start(int k, int l)
// =========================================================
{
	int i,j;
	A_M(0,0) = 0;
	A_H(0,0) = 0;
	A_P(0,0) = 0;
	A_Q(0,0) = 0;
	for (i=1; i<=m; i++)
	{
		A_P(i,0) = 0;
		A_Q(i,0) = 0;
		A_M(i,0) = 0;
		A_H(i,0) = 0;
	}
	for (j=1; j<=n; j++)
	{
		A_Q(0,j) = 0;
		A_P(0,j) = 0;
		A_M(0,j) = 0;
		A_H(0,j) = 0;
	}
}

// =========================================================
void pairwise_cyclic_alignment::compute_matrices(int k, int l)
// =========================================================
{
	int i,j;
	
	for (i=1; i<=m; i++)
	{
		for (j=1; j<=n; j++)
		{
			A_P(i,j) = max(
				A_H(i-1,j) + gamma,
				A_P(i-1,j) + alpha
			);
			A_Q(i,j) = max(
				A_H(i,j-1) + gamma,
				A_Q(i,j-1) + alpha
			);
			A_M(i,j) = A_H(i-1,j-1) + match_score(X[ (i-1+k) % m ],Y[ (j-1+l) % n ]);
			A_H(i,j) = max( A_M(i,j) , A_P(i,j) , A_Q(i,j));
		}
	}
}

// =========================================================
void pairwise_cyclic_alignment::compute_weighted_matrices(int k, int l)
// =========================================================
{
	int i,j;

	float_list X_normalized_line_weights = X.get_normalized_weights();
	float_list Y_normalized_line_weights = Y.get_normalized_weights();
//	std::cout<<"weights X: ";
//	float_list::iterator Xl_it = X_normalized_line_weights.begin();
//	float_list::iterator Xl_end = X_normalized_line_weights.end();
//	for (; Xl_it!=Xl_end; ++Xl_it)
//	{
//		std::cout<<*Xl_it<<"; ";
//		float dbx = *Xl_it;
//	}	
//	std::cout<<"\n";
//	std::cout<<"weights Y: ";
//	float_list::iterator Yl_it = Y_normalized_line_weights.begin();
//	float_list::iterator Yl_end = Y_normalized_line_weights.end();
//	for (; Yl_it!=Yl_end; ++Yl_it)
//	{
//		std::cout<<*Yl_it<<"; ";
//		float dbx = *Yl_it;
//	}
//	std::cout<<"\n";
	
	for (i=1; i<=m; i++)
	{
		for (j=1; j<=n; j++)
		{
			A_P(i,j) = max(
				A_H(i-1,j) + gamma,
				A_P(i-1,j) + alpha
			);
			A_Q(i,j) = max(
				A_H(i,j-1) + gamma,
				A_Q(i,j-1) + alpha
			);
			A_M(i,j) = A_H(i-1,j-1) + match_score(X[ (i-1+k) % m ],Y[ (j-1+l) % n ],X_normalized_line_weights,Y_normalized_line_weights);
			A_H(i,j) = max( A_M(i,j) , A_P(i,j) , A_Q(i,j));
		}
	}
}

// =========================================================
void pairwise_cyclic_alignment::compute_local_matrices(int k, int l)
// =========================================================
{
	int i,j;
	
	for (i=1; i<=m; i++)
	{
		for (j=1; j<=n; j++)
		{
			A_P(i,j) = max(
				A_H(i-1,j) + local_gamma,
				A_P(i-1,j) + local_alpha,
				0
			);
			A_Q(i,j) = max(
				A_H(i,j-1) + local_gamma,
				A_Q(i,j-1) + local_alpha,
				0
			);
			A_M(i,j) = max( A_H(i-1,j-1) + local_match_score(X[ (i-1+k) % m ],Y[ (j-1+l) % n ]) , 0 );
			A_H(i,j) = max( A_M(i,j) , A_P(i,j) , A_Q(i,j));
		}
	}
}

// =========================================================
void pairwise_cyclic_alignment::set_output_width(int header_w, int body_w)
// =========================================================
{
	header_width = header_w;
	body_width = body_w;
}

// =========================================================
void pairwise_cyclic_alignment::set_gap_costs(int open, int ext)
// =========================================================
{
	gamma = open;
	alpha = ext;
	if (gamma>0)
		gamma = -gamma;
	if (alpha>0)
		alpha = -alpha;
	
	local_gamma = (score_type)((float)(gamma * single_local_match_score)
		         / (float)(single_local_match_score - single_local_mismatch_score));
	local_alpha = -max( 1 ,
		     (score_type)((float)(-alpha * single_local_match_score)
			/ (float)(single_local_match_score - single_local_mismatch_score) ));
}



std::ostream& operator<<(std::ostream& os, pairwise_cyclic_alignment& A)
{
	A.result.set_output_width(A.header_width,A.body_width);
	os<<"score: "<<A.trace_score<<"\n\n"<<A.result<<"\n";	
	return os;
}

